; An example of Direct3D 8 programming with MASM.

; How to compile
;
; ml /c /coff /Ic:\masm\include vector.asm
; ml /c /coff /Ic:\masm\include matrix.asm
; ml /c /coff /Ic:\masm\include d3dsetup.asm
; ml /c /coff /Ic:\masm\include d3d8_ex2.asm
; link /SUBSYSTEM:WINDOWS /MACHINE:IX86 /OPT:NOREF /ENTRY:Start /OUT:d3d8_ex1.exe d3d8_ex1.obj d3dsetup.obj matrix.obj vector.obj
;
; Change the path in the /I option to whatever folder matches your system.

; /Mic, 2003
; stabmaster_@hotmail.com
; http://www.cyd.liu.se/~micol972/site/


.486
.model flat,stdcall
option casemap:none


include .\d3d8.inc
include .\macros.inc
include .\matrix.inc
include .\d3dsetup.inc
include .\cube.inc
include kernel32.inc
include user32.inc
include gdi32.inc

; You may have to change these if you keep MASM somewhere else.
includelib \masm\lib\kernel32.lib
includelib \masm\lib\user32.lib
includelib \masm\lib\gdi32.lib



XDIM	EQU 640
YDIM	EQU 480



.data
	; This is defined in d3dsetup.asm and will contain a pointer to a IDirect3DDevice8
	; interface after CreateD3DDevice has been called.
	EXTERN lpD3DDevice:DWORD

	winCaption	db 'DirectX 8 for MASM',0
	myClassName	db 'DX8MASMCLASS',0
	szTextureName	db 'mad040.raw',0
	
	; These will be used to transform the world matrix.
	xtrans		REAL4 -0.5
	ytrans		REAL4 -0.5
	ztrans		REAL4 -3.5
	
	fov		REAL4 1.0472	; Field of vision: 60 * PI / 180.
	aspect		REAL4 0.75	; Aspect ration: height/width.
	zNear		REAL4 1.0	; Distance to near clipping plane.
	zFar		REAL4 100.0	; Distance to far clipping plane.

	clearColor	dd 0
	clearDepth	REAL4 1.0

	; The lightsource. It's a point-light positioned at {0, 0, 15} with range 50 and
	; a constant attenuation of 1.
  	light		D3DLIGHT8 <D3DLIGHT_POINT,\		
  	                           <0.7,0.7,0.7,0.0>, <0.9,0.9,0.0,0.0>, <0.1,0.1,0.1,0.0>,\		
	                           <0.0,0.0,15.0>, <0.0,0.0,0.0>,\		
	                           50.0, 0.0,\			
	                           1.0, 0.0, 0.0,\			
	                           0.0, 0.0>				

	; And here's the material we'll use. It's on the form diffuse, specular, ambient, emissive, power.
	material	D3DMATERIAL8 <<1.0,1.0,1.0,0.0>, <1.0,1.0,1.0,0.0>, <1.0,1.0,1.0,0.0>,\
	                              <0.0,0.0,0.0,0.0>, 0.0>

	cameraPos	D3DXVECTOR <0.0, 0.0,15.0>	; Position of camera (or "eye").
	lookAt		D3DXVECTOR <0.0, 0.0, 0.0>	; The vector the camera will be looking along.
	upVector	D3DXVECTOR <0.0, 1.0, 0.0>	; The world up-vector.

	theta		REAL4 0.0
	d_theta		REAL4 0.020
	
.data?
	hWindow		dd ?
	lpTexture	dd ?
	lpVertexBuffer	dd ?
	lpVertexBufferData dd ?
	
	viewMatrix	D3DXMATRIX <?>
	projMatrix	D3DXMATRIX <?>
	worldMatrix	D3DXMATRIX <?>
	tempMatrix	D3DXMATRIX <?>
	rotMatrix	D3DXMATRIX <?>	
	d3dlr		D3DLOCKED_RECT <?>	

	
.code



; This function clears the background and draws a shaded triangle.
Draw PROC
	pusha
	
	COINVOKE lpD3DDevice,IDirect3DDevice8,Clear, 0, NULL, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER,\
	                                             clearColor, clearDepth, 0

	; Beginning of scene.	
	COINVOKE lpD3DDevice,IDirect3DDevice8,BeginScene

		; Rotate 'theta' radians about all three axes.
		invoke _D3DXMatrixRotationX,ADDR worldMatrix,theta
		invoke _D3DXMatrixRotationY,ADDR tempMatrix,theta
		invoke _D3DXMatrixMultiply,ADDR worldMatrix,ADDR worldMatrix,ADDR tempMatrix
		invoke _D3DXMatrixRotationZ,ADDR tempMatrix,theta
		invoke _D3DXMatrixMultiply,ADDR worldMatrix,ADDR worldMatrix,ADDR tempMatrix

		; Finally do the translation.		
		invoke _D3DXMatrixTranslation, ADDR tempMatrix, xtrans,ytrans, ztrans
		invoke _D3DXMatrixMultiply,ADDR worldMatrix,ADDR worldMatrix,ADDR tempMatrix

		; And set this to the current world matrix.		
		COINVOKE lpD3DDevice,IDirect3DDevice8,SetTransform, D3DTS_WORLD, ADDR worldMatrix		
	
		; Set the stream source to our vertex buffer.
		COINVOKE lpD3DDevice,IDirect3DDevice8,SetStreamSource, 0, lpVertexBuffer, sizeof CUSTOM_VERTEX

		; Draw 12 triangles (6 faces of a cube).
		COINVOKE lpD3DDevice,IDirect3DDevice8,DrawPrimitive, D3DPT_TRIANGLELIST, 0, 12

	; End of scene.
	COINVOKE lpD3DDevice,IDirect3DDevice8,EndScene

	; Swap the front and back buffers.
	COINVOKE lpD3DDevice,IDirect3DDevice8,Present, NULL, NULL, NULL, NULL

	fld dword ptr [theta]
	fadd dword ptr [d_theta]
	fstp dword ptr [theta]

	popa
	ret
Draw ENDP



InitTexture PROC
	LOCAL fileHandle:DWORD,lpRow:DWORD,dummy:DWORD
	
	pusha

	; Create a texture 256x256 elements large.	
	COINVOKE lpD3DDevice,IDirect3DDevice8,CreateTexture, 256, 256, 0, 0,\
					      D3DFMT_A8R8G8B8, D3DPOOL_MANAGED, ADDR lpTexture
	.if eax<0
		MSGBOX "Failed to create texture"
		popa
		return FALSE
	.endif

	; Lock the texture rect so that we can write to it.	
	COINVOKE lpTexture,IDirect3DTexture8,LockRect,0,ADDR d3dlr,0,0
	.if d3dlr.pBits == NULL
		MSGBOX "Failed to lock texture"
		popa
		return FALSE
	.endif

	; Open the texture file
	invoke CreateFile,ADDR szTextureName, GENERIC_READ, 0, NULL, OPEN_EXISTING,\
                    	  FILE_ATTRIBUTE_NORMAL or FILE_FLAG_SEQUENTIAL_SCAN,NULL
 	.if eax == INVALID_HANDLE_VALUE
    		MSGBOX "Missing file: mad040.raw"
    		return FALSE
    	.endif
    	mov fileHandle,eax

	;---------------------------------------------------------------------------------------
	; The texture file is BGR format and we want it in ARGB format so we'll convert it here.
	
	; Allocate memory for one row of image data.
	invoke LocalAlloc,LPTR,768
	.if eax == NULL
		MSGBOX "Out of memory"
		return NULL
	.endif
	mov lpRow,eax

	mov edi,d3dlr.pBits
	mov edx,d3dlr.Pitch

	mov ecx,256
	mov ebx,0
	@@upload_texture_y:
		pusha
		invoke ReadFile, fileHandle, lpRow, 768, ADDR dummy, NULL
		popa
		push edi
		push ecx
		mov esi,lpRow
		mov ecx,256
		@@upload_texture_x:
			mov al,[esi]
			mov [edi+2],al
			mov al,[esi+1]
			mov [edi+1],al
			mov al,[esi+2]
			mov [edi],al
			; Set alpha to FFh 
			mov al,0FFh
			mov [edi+3],al
			add esi,3
			add edi,4
			dec ecx
		jnz @@upload_texture_x
		pop ecx
		pop edi
		add edi,edx
		dec ecx
	jnz @@upload_texture_y

	invoke LocalFree,lpRow	
	;---------------------------------------------------------------------------------------

	invoke CloseHandle,fileHandle
	
	COINVOKE lpTexture,IDirect3DTexture8,UnlockRect,0

	; Set our created texture to the current one.
	COINVOKE lpD3DDevice,IDirect3DDevice8,SetTexture, 0, lpTexture

	; Set up the texture parameters.
	COINVOKE lpD3DDevice,IDirect3DDevice8,SetTextureStageState, 0, D3DTSS_MINFILTER, D3DTEXF_LINEAR
	COINVOKE lpD3DDevice,IDirect3DDevice8,SetTextureStageState, 0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR
	COINVOKE lpD3DDevice,IDirect3DDevice8,SetTextureStageState, 0, D3DTSS_COLOROP, D3DTOP_MODULATE4X
	COINVOKE lpD3DDevice,IDirect3DDevice8,SetTextureStageState, 0, D3DTSS_COLORARG1, D3DTA_TEXTURE
	COINVOKE lpD3DDevice,IDirect3DDevice8,SetTextureStageState, 0, D3DTSS_COLORARG2, D3DTA_CURRENT

	popa
	return lpTexture
InitTexture ENDP

 


; This functions creates a Direct3D device and sets some initial parameters.
Init PROC
	LOCAL temp:DWORD
	
	pusha
	
	; Create a D3D device. The parameters are:
	;  * Handle of owner window
	;  * Width of owner window
	;  * Height of owner window
	;  * Desired bits per pixel
	;  * Fullscreen flag (true/false)
	;
	invoke CreateD3DDevice,hWindow,XDIM,YDIM,32,0
	.if eax == FALSE
		invoke ExitProcess,0
		popa
		ret
	.endif

	; Set the culling mode to clockwise defined backfaces.
	COINVOKE lpD3DDevice,IDirect3DDevice8,SetRenderState, D3DRS_CULLMODE, D3DCULL_CW

	; Set the vertex format.
	COINVOKE lpD3DDevice,IDirect3DDevice8,SetVertexShader, D3DFVF_XYZ + D3DFVF_NORMAL + D3DFVF_TEX1

	; Create a vertex buffer
	mov eax,36
	mov ebx,sizeof CUSTOM_VERTEX
	mul ebx
	mov temp,eax
	COINVOKE lpD3DDevice,IDirect3DDevice8,CreateVertexBuffer, temp, 0,\
	                                                          D3DFVF_XYZ + D3DFVF_NORMAL + D3DFVF_TEX1,\
	                                                          D3DPOOL_MANAGED, ADDR lpVertexBuffer
	.if lpVertexBuffer == NULL
		MSGBOX "Failed to create vertex buffer"
		invoke ExitProcess,0
	.endif


	; Lock the vertex buffer.
	COINVOKE lpVertexBuffer,IDirect3DVertexBuffer8,Lock_,0, 0, ADDR lpVertexBufferData, NULL
	
	; Populate the vertex buffer with the cube data.
	mov esi,OFFSET vertexes
	mov edi,lpVertexBufferData
	mov ecx,temp
	shr ecx,2
	rep movsd

	; Unlock the vertex buffer
	COINVOKE lpVertexBuffer,IDirect3DVertexBuffer8,Unlock
	
        invoke InitTexture
        .if eax == NULL
        	invoke ExitProcess,0
        .endif
         
	
	; Set up a view matrix.
	invoke _D3DXMatrixLookAtRH, ADDR viewMatrix, ADDR cameraPos, ADDR lookAt, ADDR upVector
	COINVOKE lpD3DDevice,IDirect3DDevice8,SetTransform, D3DTS_VIEW, ADDR viewMatrix
	
	; Set up a projection matrix. 
	invoke _D3DXMatrixPerspectiveFovRH, ADDR projMatrix, fov, aspect, zNear, zFar
	COINVOKE lpD3DDevice,IDirect3DDevice8,SetTransform, D3DTS_PROJECTION, ADDR projMatrix

	; Disable Z-buffering.
	COINVOKE lpD3DDevice,IDirect3DDevice8,SetRenderState, D3DRS_ZENABLE, D3DZB_FALSE
	
	; Enable lighting.
	COINVOKE lpD3DDevice,IDirect3DDevice8,SetRenderState, D3DRS_LIGHTING, TRUE
	COINVOKE lpD3DDevice,IDirect3DDevice8,LightEnable, 0, TRUE		

	; Set the light and material parameters.
	COINVOKE lpD3DDevice,IDirect3DDevice8,SetLight, 0, ADDR light
	COINVOKE lpD3DDevice,IDirect3DDevice8,SetMaterial, ADDR material

	popa
	ret
Init ENDP




WndProc PROC hWnd:HWND,uMsg:UINT,wParam:WPARAM,lParam:LPARAM
	LOCAL crect:RECT
	LOCAL ps:PAINTSTRUCT

 	mov eax,uMsg

 	.if eax==WM_PAINT
    		invoke BeginPaint,hWnd,ADDR ps

		; Draw another frame    		
    		invoke Draw
    		
    		invoke EndPaint,hWnd,ADDR ps
    		xor eax,eax
  
  	.elseif eax==WM_KEYDOWN
    		.if wParam==VK_ESCAPE
      			invoke PostMessage,hWnd,WM_CLOSE,NULL,NULL
    		.endif
		xor eax,eax
		
  	.elseif eax==WM_DESTROY
    		invoke PostQuitMessage,0
 
	.else
 		invoke DefWindowProc,hWnd,uMsg,wParam,lParam
 	.endif
 
 	ret
WndProc ENDP




WinMain PROC hInst:DWORD,prevInstance:DWORD,cmdlinePtr:DWORD,cmdShow:DWORD
	LOCAL wc:WNDCLASSEX
	LOCAL msg:MSG

 	lea edi,wc
 	xor eax,eax
 	mov ecx,(sizeof WNDCLASSEX)/4
 	rep stosd
 	mov eax,hInst
  	mov wc.hInstance,eax
 	mov wc.cbSize, sizeof WNDCLASSEX
 	mov wc.style,CS_OWNDC or CS_HREDRAW or CS_VREDRAW
 	mov wc.lpszClassName,offset myClassName
 	mov wc.lpfnWndProc,offset WndProc
 	invoke LoadCursor,NULL,IDC_ARROW
 	mov wc.hCursor,eax
 	mov wc.hbrBackground,0

 	invoke RegisterClassEx,ADDR wc

 	sub eax,eax 
 	invoke CreateWindowEx, NULL,ADDR myClassName,ADDR winCaption,\
 	                       WS_OVERLAPPEDWINDOW OR WS_CLIPCHILDREN OR WS_CLIPSIBLINGS,\
 	                       80, 80, XDIM, YDIM, eax, eax, hInst, eax
 	mov hWindow,eax

	; Initialize
	invoke Init

 	invoke ShowWindow,eax,SW_SHOWDEFAULT
 	invoke UpdateWindow,hWindow  

 	@@winmainmsgloop:
  		invoke InvalidateRect,hWindow,NULL,FALSE 
  		sub eax,eax
  		lea edi,msg
  		invoke GetMessage,edi,eax,eax,eax
  		test eax,eax
  		jz @@quitmsgposted
  		push edi
  		push edi
  		call TranslateMessage
  		call DispatchMessage
  	jmp @@winmainmsgloop
 
 	@@quitmsgposted:

	COINVOKE lpVertexBuffer,IDirect3DVertexBuffer8,Release

	; Destroy the D3D device.  
  	invoke DestroyD3DDevice
  	
  	mov eax,msg.wParam
 	Abort::
  	invoke ExitProcess,eax
  
  	ret
WinMain ENDP



Start:

invoke GetModuleHandle,NULL 
invoke WinMain,eax,NULL,NULL,NULL

END Start

